@@ -82,6 +82,8 @@ gboolean no_daemon = FALSE;	/* Do not daemonize */
 GdmConnection *fifoconn = NULL; /* Fifo connection */
 GdmConnection *unixconn = NULL; /* UNIX Socket connection */
 
+char *gdm_charset = NULL;
+
 /* True if the server that was run was in actuallity not specified in the
  * config file.  That is if xdmcp was disabled and no local servers were
  * defined.  If the user kills all his local servers by mistake and keeps
@@ -801,7 +803,7 @@ gdm_final_cleanup (void)
 	if (extra_process > 1) {
 		/* we sigterm extra processes, and we
 		 * don't wait */
-		kill (extra_process, SIGTERM);
+		kill (-(extra_process), SIGTERM);
 		extra_process = 0;
 	}
 
@@ -829,7 +831,7 @@ gdm_final_cleanup (void)
 		 * because X is stupid and full of races and will hang my
 		 * keyboard if I don't */
 		if (li != list)
-			sleep (2);
+			gdm_sleep_no_signal (2);
 		gdm_display_unmanage (d);
 	}
 	g_slist_free (list);
@@ -919,7 +921,7 @@ deal_with_x_crashes (GdmDisplay *d)
 		    gnome_setenv ("SBINDIR", EXPANDED_SBINDIR, TRUE);
 
 		    /* To enable gettext stuff in the script */
-		    gnome_setenv ("TEXTDOMAIN", PACKAGE, TRUE);
+		    gnome_setenv ("TEXTDOMAIN", GETTEXT_PACKAGE, TRUE);
 		    gnome_setenv ("TEXTDOMAINDIR", GNOMELOCALEDIR, TRUE);
 
 		    execv (argv[0], argv);
@@ -976,6 +978,30 @@ deal_with_x_crashes (GdmDisplay *d)
     return FALSE;
 }
 
+static void
+suspend_machine (void)
+{
+	if (GdmSuspendReal != NULL &&
+	    fork () == 0) {
+		char **argv;
+		if (GdmXdmcp)
+			gdm_xdmcp_close ();
+		/* In the child setup empty mask and set all signals to
+		 * default values */
+		gdm_unset_signals ();
+
+		/* Also make a new process group */
+		setsid ();
+
+		chdir ("/");
+
+		argv = ve_split (GdmSuspendReal);
+		execv (argv[0], argv);
+		/* FIXME: what about fail */
+		_exit (1);
+	}
+}
+
 static gboolean 
 gdm_cleanup_children (void)
 {
@@ -1140,14 +1166,7 @@ start_autopsy:
 
     case DISPLAY_SUSPEND:	/* Suspend machine */
 	gdm_info (_("gdm_child_action: Master suspending..."));
-
-	gdm_final_cleanup ();
-	chdir ("/");
-
-	argv = ve_split (GdmSuspendReal);
-	execv (argv[0], argv);
-
-	gdm_error (_("gdm_child_action: Suspend failed: %s"), strerror (errno));
+	suspend_machine ();
 
 	status = DISPLAY_REMANAGE;
 	goto start_autopsy;
@@ -1407,6 +1426,7 @@ main (int argc, char *argv[])
     FILE *pf;
     poptContext ctx;
     int nextopt;
+    const char *charset;
 
     gdm_main_pid = getpid ();
 
@@ -1414,6 +1434,7 @@ main (int argc, char *argv[])
     ensure_desc_012 ();
 
     store_argv (argc, argv);
+    gdm_saveenv ();
 
     bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
     textdomain (GETTEXT_PACKAGE);
@@ -1447,6 +1468,9 @@ main (int argc, char *argv[])
     main_loop = g_main_loop_new (NULL, FALSE);
     openlog ("gdm", LOG_PID, LOG_DAEMON);
 
+    if ( ! g_get_charset (&charset)) {
+	    gdm_charset = g_strdup (charset);
+    }
 
     /* Parse configuration file */
     gdm_config_parse();
