@@ -37,6 +37,7 @@
 #ifdef HAVE_SYS_SOCKIO_H
 #include <sys/sockio.h>
 #endif
+#include <sys/resource.h>
 
 #include <vicious.h>
 
@@ -46,6 +47,7 @@
 
 /* Configuration option variables */
 extern gchar *GdmPidFile;
+extern gchar *GdmLocalNoPasswordUsers;
 extern gboolean GdmDebug;
 extern GSList *displays;
 extern int gdm_xdmcpfd;
@@ -54,9 +56,12 @@ extern char **environ;
 
 extern pid_t extra_process;
 extern int extra_status;
+extern pid_t gdm_main_pid;
 extern gboolean preserve_ld_vars;
 extern int gdm_in_signal;
 
+extern char *gdm_charset;
+
 static void
 do_syslog (int type, const char *s)
 {
@@ -92,13 +97,28 @@ gdm_fail (const gchar *format, ...)
 
     /* Log to both syslog and stderr */
     do_syslog (LOG_CRIT, s);
-    fprintf (stderr, "%s\n", s);
-    fflush (stderr);
+    if (getpid () == gdm_main_pid) {
+	    gdm_fdprintf (2, "%s\n", s);
+    }
 
     g_free (s);
 
-    if (GdmPidFile != NULL)
-	    unlink (GdmPidFile);
+    /* If main process do final cleanup to kill all processes */
+    if (getpid () == gdm_main_pid) {
+	    gdm_final_cleanup ();
+    } else if ( ! gdm_slave_final_cleanup ()) {
+	    /* If we weren't even a slave do some random cleanup only */
+	    /* FIXME: is this all fine? */
+	    gdm_sigchld_block_push ();
+	    if (extra_process > 1 && extra_process != getpid ()) {
+		    /* we sigterm extra processes, and we
+		     * don't wait */
+		    kill (-(extra_process), SIGTERM);
+		    extra_process = 0;
+	    }
+	    gdm_sigchld_block_pop ();
+    }
+
     closelog ();
 
     /* Slow down respawning if we're started from init */
@@ -253,6 +273,36 @@ gdm_clearenv (void)
 	g_list_free (envs);
 }
 
+static GList *stored_env = NULL;
+
+void
+gdm_saveenv (void)
+{
+	int i;
+
+	g_list_foreach (stored_env, (GFunc)g_free, NULL);
+	g_list_free (stored_env);
+	stored_env = NULL;
+
+	for (i = 0; environ[i] != NULL; i++) {
+		char *env = environ[i];
+		stored_env = g_list_prepend (stored_env, g_strdup (env));
+	}
+}
+
+/* leaks */
+void
+gdm_restoreenv (void)
+{
+	GList *li;
+
+	gnome_clearenv ();
+
+	for (li = stored_env; li != NULL; li = li->next) {
+		putenv (g_strdup (li->data));
+	}
+}
+
 /* Evil function to figure out which display number is free */
 int
 gdm_get_free_display (int start, uid_t server_uid)
@@ -270,6 +320,7 @@ gdm_get_free_display (int start, uid_t server_uid)
 		GSList *li;
 		struct stat s;
 		char buf[256];
+		FILE *fp;
 
 		for (li = displays; li != NULL; li = li->next) {
 			GdmDisplay *dsp = li->data;
@@ -295,6 +346,32 @@ gdm_get_free_display (int start, uid_t server_uid)
 		}
 		close (sock);
 
+		/* if lock file exists and the process exists */
+		g_snprintf (buf, sizeof (buf), "/tmp/.X%d-lock", i);
+		if (stat (buf, &s) == 0 &&
+		    ! S_ISREG (s.st_mode)) {
+			/* Eeeek! not a regular file?  Perhaps someone
+			   is trying to play tricks on us */
+			continue;
+		}
+		fp = fopen (buf, "r");
+		if (fp != NULL) {
+			char buf2[100];
+			if (fgets (buf2, sizeof (buf2), fp) != NULL) {
+				gulong pid;
+				if (sscanf (buf2, "%lu", &pid) == 1 &&
+				    kill (pid, 0) == 0) {
+					fclose (fp);
+					continue;
+				}
+
+			}
+			fclose (fp);
+
+			/* whack the file, it's a stale lock file */
+			unlink (buf);
+		}
+
 		/* if starting as root, we'll be able to overwrite any
 		 * stale sockets or lock files, but a user may not be
 		 * able to */
@@ -324,53 +401,62 @@ gboolean
 gdm_text_message_dialog (const char *msg)
 {
 	char *dialog; /* do we have dialog?*/
+	char *msg_quoted;
 
 	if (access (EXPANDED_SBINDIR "/gdmopen", X_OK) != 0)
 		return FALSE;
 
+	msg_quoted = g_shell_quote (msg);
+	
 	dialog = g_find_program_in_path ("dialog");
 	if (dialog == NULL)
 		dialog = g_find_program_in_path ("gdialog");
 	if (dialog == NULL)
 		dialog = g_find_program_in_path ("whiptail");
 	if (dialog != NULL) {
-		char *argv[7];
-
+		char *argv[6];
+		
 		argv[0] = EXPANDED_SBINDIR "/gdmopen";
-		argv[1] = dialog;
-		argv[2] = "--msgbox";
-		argv[3] = (char *)msg;
-		argv[4] = "11";
-		argv[5] = "70";
-		argv[6] = NULL;
+		argv[1] = "-l";
+		argv[2] = "/bin/sh";
+		argv[3] = "-c";
+		argv[4] = g_strdup_printf ("%s --msgbox %s 11 70",
+					   dialog, msg_quoted);
+		argv[5] = NULL;
 
 		/* make sure gdialog wouldn't get confused */
 		if (gdm_exec_wait (argv, TRUE /* no display */,
 				   TRUE /* de_setuid */) < 0) {
 			g_free (dialog);
+			g_free (msg_quoted);
+			g_free (argv[4]);
 			return FALSE;
 		}
 
 		g_free (dialog);
+		g_free (argv[4]);
 	} else {
-		char *argv[5];
+		char *argv[6];
 
 		argv[0] = EXPANDED_SBINDIR "/gdmopen";
-		argv[1] = "/bin/sh";
-		argv[2] = "-c";
-		argv[3] = g_strdup_printf
+		argv[1] = "-l";
+		argv[2] = "/bin/sh";
+		argv[3] = "-c";
+		argv[4] = g_strdup_printf
 			("clear ; "
-			 "echo \"%s\" ; read ; clear",
-			 msg);
-		argv[4] = NULL;
+			 "echo %s ; read ; clear",
+			 msg_quoted);
+		argv[5] = NULL;
 
 		if (gdm_exec_wait (argv, TRUE /* no display */,
 				   TRUE /* de_setuid */) < 0) {
-			g_free (argv[3]);
+			g_free (argv[4]);
+			g_free (msg_quoted);
 			return FALSE;
 		}
-		g_free (argv[3]);
+		g_free (argv[4]);
 	}
+	g_free (msg_quoted);
 	return TRUE;
 }
 
@@ -378,6 +464,7 @@ gboolean
 gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 {
 	char *dialog; /* do we have dialog?*/
+	char *msg_quoted;
 
 	if (access (EXPANDED_SBINDIR "/gdmopen", X_OK) != 0)
 		return FALSE;
@@ -385,36 +472,42 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 	if (ret != NULL)
 		*ret = FALSE;
 
+	msg_quoted = g_shell_quote (msg);
+	
 	dialog = g_find_program_in_path ("dialog");
 	if (dialog == NULL)
 		dialog = g_find_program_in_path ("gdialog");
 	if (dialog == NULL)
 		dialog = g_find_program_in_path ("whiptail");
 	if (dialog != NULL) {
-		char *argv[7];
+		char *argv[6];
 		int retint;
 
 		argv[0] = EXPANDED_SBINDIR "/gdmopen";
-		argv[1] = dialog;
-		argv[2] = "--yesno";
-		argv[3] = (char *)msg;
-		argv[4] = "11";
-		argv[5] = "70";
-		argv[6] = NULL;
+		argv[1] = "-l";
+		argv[2] = "/bin/sh";
+		argv[3] = "-c";
+		argv[4] = g_strdup_printf ("%s --yesno %s 11 70",
+					   dialog, msg_quoted);
+		argv[5] = NULL;
 
 		/* will unset DISPLAY and XAUTHORITY if they exist
 		 * so that gdialog (if used) doesn't get confused */
 		retint = gdm_exec_wait (argv, TRUE /* no display */,
 					TRUE /* de_setuid */);
 		if (retint < 0) {
+			g_free (argv[4]);
 			g_free (dialog);
+			g_free (msg_quoted);
 			return FALSE;
 		}
 
 		if (ret != NULL)
 			*ret = (retint == 0) ? TRUE : FALSE;
 
+		g_free (argv[4]);
 		g_free (dialog);
+		g_free (msg_quoted);
 
 		return TRUE;
 	} else {
@@ -422,33 +515,37 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 		int tempfd;
 		FILE *fp;
 		char buf[256];
-		char *argv[5];
+		char *argv[6];
 
 		tempfd = g_mkstemp (tempname);
-		if (tempfd < 0)
+		if (tempfd < 0) {
+			g_free (msg_quoted);
 			return FALSE;
+		}
 
 		close (tempfd);
 
 		argv[0] = EXPANDED_SBINDIR "/gdmopen";
-		argv[1] = "/bin/sh";
-		argv[2] = "-c";
-		argv[3] = g_strdup_printf
+		argv[1] = "-l";
+		argv[2] = "/bin/sh";
+		argv[3] = "-c";
+		argv[4] = g_strdup_printf
 			("clear ; "
-			 "echo \"%s\" ; echo ; echo \"%s\" ; "
+			 "echo %s ; echo ; echo \"%s\" ; "
 			 "read RETURN ; echo $RETURN > %s ; clear'",
-			 msg,
+			 msg_quoted,
 			 /* Translators, don't translate the 'y' and 'n' */
 			 _("y = Yes or n = No? >"),
 			 tempname);
-		argv[4] = NULL;
+		argv[5] = NULL;
 
 		if (gdm_exec_wait (argv, TRUE /* no display */,
 				   TRUE /* de_setuid */) < 0) {
-			g_free (argv[3]);
+			g_free (argv[4]);
+			g_free (msg_quoted);
 			return FALSE;
 		}
-		g_free (argv[3]);
+		g_free (argv[4]);
 
 		if (ret != NULL) {
 			fp = fopen (tempname, "r");
@@ -458,12 +555,14 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 					*ret = TRUE;
 				fclose (fp);
 			} else {
+				g_free (msg_quoted);
 				return FALSE;
 			}
 		}
 
 		unlink (tempname);
 
+		g_free (msg_quoted);
 		return TRUE;
 	}
 }
@@ -488,9 +587,9 @@ gdm_exec_wait (char * const *argv, gboolean no_display,
 
 		/* No error checking here - if it's messed the best response
 		 * is to ignore & try to continue */
-		open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 
 		if (de_setuid) {
 			gdm_desetuid ();
@@ -610,10 +709,23 @@ gdm_fork_extra (void)
 
 	gdm_sigterm_block_push ();
 	pid = extra_process = fork ();
+	if (pid < 0)
+		extra_process = 0;
 	gdm_sigterm_block_pop ();
 
 	gdm_sigchld_block_pop ();
 
+	if (pid == 0) {
+		/* In the child setup empty mask and set all signals to
+		 * default values */
+		gdm_unset_signals ();
+
+		/* Also make a new process group so that we may use
+		 * kill -(extra_process) to kill extra process and all it's
+		 * possible children */
+		setsid ();
+	}
+
 	return pid;
 }
 
@@ -625,7 +737,7 @@ gdm_wait_for_extra (int *status)
 	if (extra_process > 0) {
 		ve_waitpid_no_signal (extra_process, &extra_status, 0);
 	}
-	extra_process = -1;
+	extra_process = 0;
 
 	if (status != NULL)
 		*status = extra_status;
@@ -978,4 +1090,183 @@ gdm_close_all_descriptors (int from, int except)
 	}
 }
 
+gboolean
+gdm_is_a_no_password_user (const char *user)
+{
+	char **vector = NULL;
+	int i;
+
+	if (ve_string_empty (GdmLocalNoPasswordUsers) ||
+	    ve_string_empty (user) ||
+	    strcmp (user, "root") == 0)
+		return FALSE;
+
+	vector = g_strsplit (GdmLocalNoPasswordUsers, ",", -1);
+	for (i = 0; vector[i] != NULL; i++) {
+		if (strcmp (vector[i], user) == 0) {
+			g_strfreev (vector);
+			return TRUE;
+		}
+	}
+	g_strfreev (vector);
+	return FALSE;
+}
+
+int
+gdm_open_dev_null (mode_t mode)
+{
+	int ret;
+	ret = open ("/dev/null", mode);
+	if (ret < 0) {
+		gdm_fail ("Cannot open /dev/null, system on crack!");
+	}
+
+	return ret;
+}
+
+void
+gdm_unset_signals (void)
+{
+	sigset_t mask; 
+
+	sigemptyset (&mask);
+	sigprocmask (SIG_SETMASK, &mask, NULL);
+
+	signal (SIGUSR1, SIG_DFL);
+	signal (SIGUSR2, SIG_DFL);
+	signal (SIGCHLD, SIG_DFL);
+	signal (SIGINT, SIG_DFL);
+	signal (SIGTERM, SIG_DFL);
+	signal (SIGPIPE, SIG_DFL);
+	signal (SIGALRM, SIG_DFL);
+	signal (SIGHUP, SIG_DFL);
+}
+
+static char *
+ascify (const char *text)
+{
+	unsigned char *p;
+	char *t = g_strdup (text);
+	for (p = (unsigned char *)t; p != NULL && *p != '\0'; p++) {
+		if (*p > 127)
+			*p = '?';
+	}
+	return t;
+}
+
+char *
+gdm_locale_to_utf8 (const char *text)
+{
+	GIConv cd;
+	char *out;
+	GError *error = NULL;
+
+	if (gdm_charset == NULL) {
+		return g_strdup (text);
+	}
+
+	cd = g_iconv_open ("UTF-8", gdm_charset);
+	if (cd == (GIConv)(-1)) {
+		return ascify (text);
+	}
+
+	out = g_convert_with_iconv (text, -1, cd, NULL, NULL, &error);
+	g_iconv_close (cd);
+	if (out == NULL) {
+		return ascify (text);
+	}
+	return out;
+}
+
+char *
+gdm_locale_from_utf8 (const char *text)
+{
+	GIConv cd;
+	char *out;
+
+	if (gdm_charset == NULL) {
+		return g_strdup (text);
+	}
+
+	cd = g_iconv_open (gdm_charset, "UTF-8");
+	if (cd == (GIConv)(-1)) {
+		return ascify (text);
+	}
+
+	out = g_convert_with_iconv (text, -1, cd, NULL, NULL, NULL);
+	g_iconv_close (cd);
+	if (out == NULL)
+		return ascify (text);
+	return out;
+}
+
+void
+gdm_sleep_no_signal (int secs)
+{
+	time_t endtime = time (NULL)+secs;
+
+	while (secs > 0) {
+		struct timeval tv;
+		tv.tv_sec = secs;
+		tv.tv_usec = 0;
+		select (0, NULL, NULL, NULL, &tv);
+		/* don't want to use sleep since we're using alarm
+		   for pinging */
+		secs = endtime - time (NULL);
+	}
+}
+
+void
+gdm_reset_limits (void)
+{
+	struct rlimit unlim = { RLIM_INFINITY, RLIM_INFINITY };
+
+	/* Note: I don't really know which ones are really very standard
+	   and which ones are not, so I just test for them all one by one */
+
+#ifdef RLIMIT_CPU
+	setrlimit (RLIMIT_CPU, &unlim);
+#endif
+#ifdef RLIMIT_DATA
+	setrlimit (RLIMIT_DATA, &unlim);
+#endif
+#ifdef RLIMIT_FSIZE
+	setrlimit (RLIMIT_FSIZE, &unlim);
+#endif
+#ifdef RLIMIT_LOCKS
+	setrlimit (RLIMIT_LOCKS, &unlim);
+#endif
+#ifdef RLIMIT_MEMLOCK
+	setrlimit (RLIMIT_MEMLOCK, &unlim);
+#endif
+#ifdef RLIMIT_NOFILE
+	setrlimit (RLIMIT_NOFILE, &unlim);
+#endif
+#ifdef RLIMIT_OFILE
+	setrlimit (RLIMIT_OFILE, &unlim);
+#endif
+#ifdef RLIMIT_NPROC
+	setrlimit (RLIMIT_NPROC, &unlim);
+#endif
+#ifdef RLIMIT_RSS
+	setrlimit (RLIMIT_RSS, &unlim);
+#endif
+#ifdef RLIMIT_STACK
+	setrlimit (RLIMIT_STACK, &unlim);
+#endif
+#ifdef RLIMIT_CORE
+	setrlimit (RLIMIT_CORE, &unlim);
+#endif
+#ifdef RLIMIT_AS
+	setrlimit (RLIMIT_AS, &unlim);
+#endif
+#ifdef RLIMIT_VMEM
+	setrlimit (RLIMIT_VMEM, &unlim);
+#endif
+#ifdef RLIMIT_PTHREAD
+	setrlimit (RLIMIT_PTHREAD, &unlim);
+#endif
+}
+
+
 /* EOF */
